#! /usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
import numpy as np
import pandas as pd

'''
The TTS (Transcription Termination Site), TSS (Transcription Start Site) and
TU (Transcription Unit) classes allow to gather information on the positions, 
reliability and sequences of these elements.
'''

class TSS:
    """ 
    Each TTS instance has to be initialized with the following attributes:
        * pos (int.): position of the TSS
        * strand (bool.): DNA strand (True for forward strand, False for 
          complementary strand)
        * genes (list of str.): list of locus tags of genes associated with 
          the TSS
        * score (float.): TSS score
    """

    def __init__(self, pos=None, strand=None, genes=[], score=None):
        self.pos = pos 
        self.strand = strand
        self.genes = genes  
        self.promoter = {}
        self.score = score


    def add_genes(self, tags, genes_dict):
        """
        add_genes completes the genes attribute of the TSS instance
        with the genes that are given as tags and that are in the
        genes_dict.

        Args:
            tags (str.): locus tags separated by commas
            genes_dict: dictionary of shape {locus tag : Gene object}

        Example:
            >>> from GRATIOSA import Genome
            >>> g = Genome.Genome("dickeya")
            >>> g.load_annotation()
            >>> from GRATIOSA.TSS_TTS_TU import TSS
            >>> tss = TSS()
            >>> tss.add_genes("Dda3937_00001,Dda3937_00002",g.genes)
        """
        for tag in tags.strip().replace(' ', '').split(','):
            if tag != '':
                if tag in genes_dict.keys():
                    if tag not in self.genes:
                        self.genes.append(tag)
                else:
                    print(tag + " not in annotations")


    def add_promoter(self, sig, *arg, **kwargs):
        """
        add_promoter completes the promoter attribute of the TSS instance.
        Creates or complete the 'promoter' of the TSS instance with the shape {[sig]:(sites)}

        Args:
            sig (str.): sigma factor
            sites (Optional [str.]): the left and right coordinates of the -10 and -35 
                    elements with the shape "-10l,-10r,-35r,-35l"
            genes_dict (Optional [dict.]): dictionary of shape {locus tag : Gene object}

        Example:
            >>> from GRATIOSA.TSS_TTS_TU import TSS
            >>> tss = TSS(pos=4707030,strand=False)
            >>> tss.add_promoter(sig = "sigma70", sites="4707037,4707046,4707062,4707068")
            >>> tss.promoter
            {'sigma70': {'sites': (4707037, 4707046, 4707062, 4707068)}}
        """

        # sites must have the shape : [-10l,-10r,-35l,-35r] with l = left, r =
        # right
        sites = kwargs.get('sites')
        self.promoter[sig] = {}
        if sites:
            self.promoter[sig]['sites'] = tuple(
                map(int, sites.strip().replace(' ', '').split(',')))


    def add_prom_elements(self, gen_seq, gen_seqcompl, shift=0, prom=[0, 0]):
        '''
        add_prom_elements extracts sequences of the different promoter 
        elements (spacer, -10, -35, discriminator, region around TSS) based 
        on -35 and -10 coordinates that are included in the promoter attribute.

        Completes the 'promoter' attribute of the TSS instance:
                Before using this method, this attribute was a dictionary of 
                shape {sigma factor: subdictionary} containing, for each 
                sigma factor, a subdictionary of shape "sites":(-10l,-10r,-35r,
                -35l)} with -10l, -10r, -35r and -35l the left and right 
                coordinates of the -10 and -35 elements. Now, this 
                subdictionary contains 4 new keys : 
                "spacer", "minus10","minus35" and "discriminator". 
                Each associated value is the sequence of the element.

        Args:
            gen_seq (str.): sequence of the forward strand
            gen_seqcompl (str.): sequence of the complementary strand
                    (WARNING: 3' -> 5')
            shift (Optional [int.]): number of nucleotides to include beyond 
                    each region on either side (Default: 0nt)
            prom (Optional [int.,int.]): region upstream and downstream TSS
                    to extract. Argument with the shape:
                    [length before TSS, length after TSS]
                    Default: [0,0] ie no sequence will be extracted around 
                    TSS.
                    
        Note: 
            If prom !=[0,0], self.promoter dictionary has also a new 
            "region" key. self.promoter['region'] returns the sequence 
            of the region chosen with the prom argument.
                

        Warning: 
            the promoter attribute has to be loaded prior using 
            this method. Use add_promoter method.
            

        Example:
            >>> from GRATIOSA import Genome
            >>> g = Genome.Genome("dickeya")
            >>> g.load_seq()
            >>> from GRATIOSA.TSS_TTS_TU import TSS
            >>> tss = TSS(pos=4707030,strand=False)
            >>> tss.add_promoter(sig = "sigma70", sites="4707037,4707046,4707062,4707068")
            >>> tss.add_prom_elements(g.seq,
            ...                       g.seqcompl,
            ...                       sig = "sigma70",
            ...                       sites="4707039,4707044,4707060,4707065",
            ...                       prom=[20,20])
            >>> tss.promoter
            {'sigma70': {'sites': (4707037, 4707046, 4707062, 4707068),
             'spacer': 'CCTCGCCCACCCTCA',
             'minus10': 'ATCATCATGA',
             'minus35': 'CCGTACC',
             'discriminator': 'ATAACC'},
             'region': 'CTCAATCATCATGAATAACCCCCCTCCTTGTGTCTTTCTTA'}
        '''
        if prom != [0, 0]:
            if self.strand:
                self.promoter['region'] = gen_seq[self.pos -
                                                  1 - prom[0]:self.pos + prom[1] + shift]

            elif self.strand == False:
                self.promoter['region'] = gen_seqcompl[self.pos -
                                                       prom[1] - 1 - shift:self.pos + prom[0] + shift][::-1]

        try:

            for sig in self.promoter.keys():  # for each sigma factor
                try:
                    sites_pos = self.promoter[sig]['sites']
                    if self.strand:
                        self.promoter[sig]['spacer'] = gen_seq[sites_pos[3] - shift:sites_pos[0] - 1 + shift]
                        self.promoter[sig]['minus10'] = gen_seq[sites_pos[0] - 1 - shift:sites_pos[1] + shift]
                        self.promoter[sig]['minus35'] = gen_seq[sites_pos[2] - 1 - shift:sites_pos[3] + shift]
                        self.promoter[sig]['discriminator'] = gen_seq[sites_pos[1] - shift:self.pos - 1 + shift]
                    elif not self.strand:
                        self.promoter[sig]['spacer'] = gen_seqcompl[sites_pos[1] - shift:sites_pos[2] - 1 + shift][::-1]
                        self.promoter[sig]['minus10'] = gen_seqcompl[sites_pos[0] - 1 - shift:sites_pos[1] + shift][::-1]
                        self.promoter[sig]['minus35'] = gen_seqcompl[sites_pos[2] - 1 - shift:sites_pos[3] + shift][::-1]
                        self.promoter[sig]['discriminator'] = gen_seqcompl[self.pos - shift:sites_pos[0] - 1 + shift][::-1]
                except Exception as e:  # sigma factor without site coordinates or invalid
                    pass

        except Exception as e:
            print('Error for TSS', self.pos, ':', e)


class TTS:
    """
    Each TTS instance has to be initialized with the following attributes:
        * left (int.) and right (int.): TTS coordinates (does not take into
          account the strand, ie right > left)
        * start (int.) and end (int.): positions of the beginning and the 
          end of the TTS
        * strand (bool.): DNA strand (True for forward strand, False for 
          complementary strand)
        * rho_dpdt (bool.): rho dependency of the TTS
        * genes (list of str.): list of locus tags of genes associated with 
          the TTS
        * score (float.): TSS score
    """

    def __init__(self, start=None, end=None, left=None, right=None,
                 strand=None, rho_dpdt=None, genes=[], seq="", score=None):
        self.start = start
        self.end = end
        self.left = left
        self.right = right
        self.strand = strand
        self.rho_dpdt = rho_dpdt
        self.genes = genes
        self.seq = seq
        self.score = score

        if strand is not None:
            if self.left is not None and self.right is not None and (
                    self.start is None or self.stop is None):
                if self.strand:
                    self.start = self.left
                    self.end = self.right
                elif not self.strand:
                    self.start = self.right
                    self.end = self.left
            elif self.start is not None and self.end is not None and (self.left is None or self.right is None):
                if self.strand:
                    self.left = self.start
                    self.right = self.end
                elif not self.strand:
                    self.right = self.start
                    self.left = self.end


    def add_genes(self, tags, genes_dict):
        """
        add_genes completes the genes attribute of the TTS instance
        with the genes that are given as tags and that are in the genes_dict.

        Args:
            tags (str.): locus tags separated by commas
            genes_dict: dictionary of shape {locus tag : Gene object}

        Example:
            >>> from GRATIOSA import Genome
            >>> g = Genome.Genome("dickeya")
            >>> g.load_annotation()
            >>> from GRATIOSA.TSS_TTS_TU import TTS
            >>> tts = TTS()
            >>> tts.add_genes("Dda3937_00001,Dda3937_00002",g.genes)
            >>> tts.genes
            ["Dda3937_00001,Dda3937_00002"]
        """
        for tag in tags.strip().replace(' ', '').split(','):
            if tag != '':
                if tag in genes_dict.keys():
                    if tag not in self.genes:
                        self.genes.append(tag)
                else:
                    print(tag + " not in annotations")


class TU:
    """
    Each TU instance has to be initialized with the following attributes:
        * left (int.) and right (int.): TU coordinates (does not take into
          account the strand, ie right > left)
        * start (int.) and end (int.): positions of the beginning and the end of the TU
        * strand (bool.): DNA strand (True for forward strand, False for 
          complementary strand)
        * genes (list of str.): list of locus tags of genes associated with the TU

    """
    def __init__(self, start=None, end=None, left=None,
                 right=None, strand=None, genes=[],expression=None):
        self.start = start
        self.end = end
        self.left = left
        self.right = right
        self.strand = strand

        if strand is not None:
            if self.left is not None and self.right is not None and (
                    self.start is None or self.stop is None):
                if self.strand:
                    self.start = self.left
                    self.end = self.right
                elif not self.strand:
                    self.start = self.right
                    self.end = self.left
            elif self.start is not None and self.end is not None and (self.left is None or self.right is None):
                if self.strand:
                    self.left = self.start
                    self.right = self.end
                elif not self.strand:
                    self.right = self.start
                    self.left = self.end
        self.genes = genes
        self.expression = expression


    def add_genes(self, tags, genes_dict):
        """
        add_genes completes the genes attribute of the TU instance
        with the genes that are given as tags and that are in the genes_dict.

        Args:
            tags (str.): locus tags separated by commas
            genes_dict: dictionary of shape {locus tag : Gene object}

        Example:
            >>> from GRATIOSA import Genome
            >>> g = Genome.Genome("dickeya")
            >>> g.load_annotation()
            >>> from GRATIOSA.TSS_TTS_TU import TU
            >>> tu = TU()
            >>> tu.add_genes("Dda3937_00001,Dda3937_00002",g.genes)
            >>> tu.genes
            ["Dda3937_00001,Dda3937_00002"]
        """
        for tag in tags.strip().replace(' ', '').split(','):
            if tag != '':
                if tag in genes_dict.keys():
                    if tag not in self.genes:
                        self.genes.append(tag)
                else:
                    print(tag + " not in annotations")


    def add_TU_expression(self, condition, TU_expr):
        """
        add_TU_expression adds an expression value associated to a new 
        condition to the expression attribute of the TU instance. If a list 
        of expressions is given in input (with the TU_expr argument), the 
        mean is computed and the mean value is added to the attribute. Elif 
        only one value is given in input, this value is directly added to the 
        attribute.

        Creates or completes the "expression" attribute of the TU instance. 
        This attribute is a dictionary of shape {condition:expression value}

        Args:
            condition (str.): name of the condition
            TU_expr (float or list of float): expression value(s)
                             
        Example:
            >>> from GRATIOSA.TSS_TTS_TU import TU
            >>> tu = TU()
            >>> tu.add_TU_expression("Control",0.24)
            >>> tu.add_TU_expression("test",[0.2,0.4])
            >>> tu.expression
            {'Control': 0.24, 'test': 0.30}
        """
        if not hasattr(self, 'expression'):
            self.expression = {}

        if isinstance(TU_expr, float):
            self.expression[condition] = TU_expr
        else:
            self.expression[condition] = np.mean(TU_expr)


    def add_TSS(self, TSS):
        """
        add_TSS attributes potential TSS to the TU object. 
        Creates or completes the 'TSS' attributes (list of tuples) of the 
        TU instance. 

        Args:
            TSS: tuple of shape (TSS position (int.),
                                 propotion of total starts (float))

        Example:
            >>> from GRATIOSA.TSS_TTS_TU import TU
            >>> tu = TU()
            >>> tu.add_TU_expression("Control",0.24)
            >>> tu.add_TU_expression("test",[0.2,0.4])
            >>> tu.expression
            {'Control': 0.24, 'test': 0.30}
        """
        if not hasattr(self, 'TSS'):
            self.TSS = []

        if TSS not in self.TSS:
            self.TSS.append(TSS)


    def add_TTS(self, TTS):
        """
        add_TTS attributes potential TTS to the TU object.
        Creates or completes the 'TTS' attributes (list of tuples) of the 
        TU instance. 

        Args:
            TTS: tuple of shape 
                    (TTS position (int.), propotion of total starts (float))
        """
        if not hasattr(self, 'TTS'):
            self.TTS = []

        if TTS not in self.TTS:
            self.TTS.append(TTS)


#! /usr/bin/env python
# -*- coding: utf-8 -*-

'''
This module allows the classification and the statistical analysis 
(enrichment test, proportion test, student test) of omics and spatial data 
loaded on the different objects created, in particular with the Genome, 
Transcriptome, Chipseq and HiC classes.
'''
import numpy as np
from statsmodels.stats import proportion, multitest
from scipy import stats
from GRATIOSA.globvar import *
from pathlib import Path
from datetime import datetime
import pandas as pd


def data_classification(data_x, data_y, class_nb, *args, **kwargs):
    '''
    Classification of data into fractions according to defined thresholds
    (using the thresholds argument), according to class sizes (using the
    class_sizes argument) or into classes of equal size (if neither the
    thresholds argument nor the class_sizes argument is specified).

    Args:
        data_x (list): list of elements such as positions or gene names
        data_y (list): list of data associated with each element such as 
                signal coverage or gene expression
        class_nb (int.): number of classes (fractions) to create
        class_names (Optionnal [list]): list of names to give to each class
                By default, each class will be named by a number between 0 
                and class_nb.
        thresholds (Optionnal [list of ints.]): Thresholds used for the classification.
                If not given, the classification will distribute the data in 
                fractions of equal sizes.
        class_sizes (Optionnal [list of ints.]): Number of elements to put in each class.
                The total must be equal to the number of  elements in data_x 
                and in data_y.

    Returns: 
        tuple: 2 dictionaries:
            * dict. of shape {class_name:list of elements}
            * dict. of shape {class_name: list of data associated with each element}

    Example:
        >>> from GRATIOSA import stat_analysis
        >>> stat_analysis.data_classification(["a","b","d","c","e"],[1,2,4,3,5], 
        ...                                   class_nb=3,thresholds = [1,5])
        ({0: ['a'], 1: ['b', 'c', 'd'], 2: []}, {0: [1], 1: [2, 3, 4], 2: []})
        >>> stat_analysis.data_classification(["a","b","d","c","e"],[1,2,4,3,5], 
        ...                                   class_nb=3,class_sizes=[1,1,3])
        ({0: ['a'], 1: ['b'], 2: ['c', 'd', 'e']}, {0: [1], 1: [2], 2: [3, 4, 5]})
        >>> stat_analysis.data_classification(["a","b","d","c","e","f"],[1,2,4,3,5,6], 
        ...                                   class_nb=3)
        ({0: ['a', 'b'], 1: ['c', 'd'], 2: ['e', 'f']}, {0: [1, 2], 1: [3, 4], 2: [5, 6]})
    '''

    class_names = kwargs.get('class_names', np.arange(class_nb))
    thresholds = kwargs.get('thresholds')
    nb_per_class = kwargs.get('class_sizes')
    data_y = list(map(float, data_y))

    # STEP 1: Checks the adequacy between arguments
    if len(class_names) != class_nb:
        class_nb = len(class_names)
        print("WARNING: the number of classes names isn't equal to the class_nb")
        print(f"WARNING: the number of classes was set to {len(class_names)}")
    if thresholds:
        if class_nb == len(thresholds) + 1:
            # If only internal thresholds were given,
            # adds the min and max to the thresholds.
            thresholds = \
                [min(data_y + thresholds)] + thresholds + [max(data_y)]
        elif class_nb != len(thresholds) - 1:
            sys.exit("Please adjust the nb_class to thresholds.")
        print("Performing the classification in the following classes:")
        for i in np.arange(class_nb):
            print(f"{class_names[i]}: {thresholds[i]} to {thresholds[i+1]}")
    if len(data_x) != len(data_y):
        sys.exit("data_x and data_y do not have the same length")
    if nb_per_class:
        if np.sum(nb_per_class) != len(data_x):
            sys.exit("The sum of nb_per_class is not equal to the length of data_x.")
        elif len(class_names) != len(nb_per_class):
            sys.exit("Please adjust the class_names to the number of classes described with class_sizes.")

    # STEP 2: Performs the classification
    # Sorts the data
    data_y, data_x = map(list, zip(*sorted(zip(data_y, data_x))))
    classif_y = {}
    classif_x = {}

    # Option 1: classification using the chosen thresholds
    if thresholds:
        t, i = 0, 0
        while t < class_nb:
            classif_y[class_names[t]] = []
            classif_x[class_names[t]] = []
            while data_y[i] <= thresholds[t + 1] and i + 1 < len(data_y):
                classif_y[class_names[t]].append(data_y[i])
                classif_x[class_names[t]].append(data_x[i])
                i += 1
            t += 1

    # Option 2: classification using the chosen class sizes
    elif nb_per_class:
        prev_class_sums = 0
        for c in np.arange(class_nb):
            classif_y[class_names[c]
                      ] = data_y[prev_class_sums:nb_per_class[c] + prev_class_sums]
            classif_x[class_names[c]
                      ] = data_x[prev_class_sums:nb_per_class[c] + prev_class_sums]
            prev_class_sums += nb_per_class[c]

    # Option 3: classification in equal-size fractions
    else:
        size_fraction = len(data_y) / class_nb
        print("Performing the classification in the following classes:")
        for c in np.arange(class_nb - 1):
            classif_y[class_names[c]] = \
                data_y[int(c * size_fraction):int((c + 1) * size_fraction)]
            classif_x[class_names[c]] = \
                data_x[int(c * size_fraction):int((c + 1) * size_fraction)]
        classif_y[class_names[c + 1]] = data_y[int((c + 1) * size_fraction):]
        classif_x[class_names[c + 1]] = data_x[int((c + 1) * size_fraction):]
        for k in classif_y.keys():
            print(f"{k}: from {min(classif_y[k])} to {max(classif_y[k])}")

    for k in classif_y.keys():
        print(f"{k}: {len(classif_x[k])} elements")

    return classif_x, classif_y


def proportion_test(dict_cats,
                    dict_features,
                    targ_features,
                    all_features="all",
                    cats="all",
                    alt_hyp="one-sided",
                    output_dir=f"{resdir}proportion_test/",
                    output_file=f"prop_test{datetime.now()}"):
    '''
    Compares the proportion of targ_features/all_features between categories.
    The proportion test is based on a normal test using the
    stats.proportion.proportions_ztest function from statsmodels.

    Args:
        dict_cats (dict.): classification of each element in a dictionary of
               shape {category:[elements]}
        dict_features (dict.): feature corresponding to each element in a
                dictionary of shape {feature:[elements]}
        targ_features (list or string): targeted feature(s)
        all_features (Optionnal [list or str.]): features that will be used to 
                calculate the proportion, including targ_features
                (default: all keys of the dict_features dictionary)
        cats (Optionnal [list]): list of categories to compare
                (default: all keys of dict_cats)
        alt_hyp (Optionnal ["two-sided" or "one-sided"]): alternative hypothesis.
                If "one-sided" is chosen, both one-sided tests will be 
                performed with the statsmodels.stats.proportions_ztest 
                function and the smaller p-value will be kept. 
                See statsmodels documentation for more details.
                (default: "one-sided")
        output_dir (Optionnal [str.]): output directory
        output_file (Optionnal [str.]): output filename

    Returns:
        Dictionary: dict. of shape {"categories":cats, "proportions":prop,
        "confidence intervals":(ci0,ci1), "p-values":{(cat1,cat2):pval}

    Note:
        Test results are saved in the output_file

    Example:
        >>> from GRATIOSA import stat_analysis
        >>> dict_cats = {"borders":["A","B","G","K","L"],
        ...              "loops":["C","D","F","H","J"],
        ...              "None":["E","I"]}
        >>> data = {"act":["A","B","C","F","H"],
        ...         "rep":["D","G","L","I"],
        ...         "None":["J","K"],"NA":["E"]}
        >>> res = stat_analysis.proportion_test(dict_cats,data,"act",
        ...                                     all_features=["act","rep"],
        ...                                     alt_hyp="two-sided")
        >>> res["categories"]
        ['borders', 'loops', 'None']
        >>> res["proportions"]
        [0.5, 0.75, 0.0]
        >>> res["confidence intervals"]
        [(0.010009, 0.98999),(0.32565, 1.0),(0.0, 0.0)]
        >>> res["p-values"]
        {'borders-loops': 0.4652088184521418,
         'borders-None': 0.3613104285261787,
         'loops-None': 0.17090352023079747}
    '''
    # Formats the objects given as arguments as lists containing no duplicates.
    if isinstance(targ_features, str):
        targ_features = [targ_features]
    else:
        targ_features = list(dict.fromkeys(targ_features))

    if isinstance(all_features, str):
        if all_features == "all":
            all_features = set(dict_features.keys())
    else:
        all_features = list(dict.fromkeys(all_features))

    if cats == "all":
        cats = list(dict_cats.keys())
    else:
        cats = list(dict.fromkeys(cats))

    # For each category, creates a dictionnary containing:
    #   one list of elements associated to at least one feature of all_features
    #   one list of elements associated to at least one feature of tar_features
    targ_feat_elements, all_feat_elements = [], []
    for f in dict_features.keys():
        if f in targ_features:
            targ_feat_elements.extend(dict_features[f])
        if f in all_features:
            all_feat_elements.extend(dict_features[f])
    list_per_cat = {}
    for c in cats:
        list_per_cat[c] = {}
        list_per_cat[c]["tar"] = set(dict_cats[c]) & set(targ_feat_elements)
        list_per_cat[c]["all"] = set(dict_cats[c]) & set(all_feat_elements)

    # For each category, computes the proportion between the number of elements
    # in targ_features and the number in all_features and its confidence interval.
    # Saves the results in a .txt file.
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    results = open(f'{output_dir}{output_file}.txt', 'w')
    results.write(f"Targeted_features: {targ_features}\n")
    results.write(f"All_features: {all_features}\n\n")
    results.write(f"Category\tTargeted_feature\tAll_features\tProportion\t"
                  "Confidence interval\n")
    prop, cim = [], []
    for c in cats:
        try:
            nb_A = len(list_per_cat[c]["tar"])
            nb_tot = len(list_per_cat[c]["all"])
            print(f"{c}: {nb_tot} elements")
            pexp = float(nb_A / nb_tot)
            std = np.array(stats.binom.std(nb_tot, pexp, loc=0)) / nb_tot
            ci = proportion.proportion_confint(nb_A, nb_tot, alpha=0.05,
                                               method='normal')
            ci0 = max(ci[0], 0)
            ci1 = min(ci[1], 1)
            cim.append((ci0, ci1))
            prop.append(pexp)
            results.write(f"{c}\t{nb_A}\t{nb_tot}\t{pexp}\t{(ci0,ci1)}\n")
        except Exception as ex:
            print(f"Error with {c}: {ex}")
    results.write("\n")

    # Tests pairwise if the proportions between categories are significantly
    # differents. Performs one-sided or two-sided proportions test depending 
    # on the "alt_hyp" argument.
    comp = []
    pvals = {}
    for i in np.arange(len(cats)):
        for j in np.arange(i + 1, len(cats)):
            comp.append([cats[i], cats[j]])
    for cat1, cat2 in comp:
        s1 = len(list_per_cat[cat1]["tar"])
        s2 = len(list_per_cat[cat2]["tar"])
        n1 = s1 + len(list_per_cat[cat1]["all"])
        n2 = s2 + len(list_per_cat[cat2]["all"])
        if alt_hyp == "one-sided":
            stat1, pval1 = proportion.proportions_ztest([s1, s2], [n1, n2],
                                                        alternative='larger')
            stat2, pval2 = proportion.proportions_ztest([s1, s2], [n1, n2],
                                                        alternative='smaller')
            pval = min(pval1, pval2)
            results.write(
                f"One-sided test\tp-value between {cat1} and {cat2}\n")
            results.write(f"Larger\t{str(pval1)}\n")
            results.write(f"Smaller\t{str(pval2)}\n")
            print(f"One-sided test\tp-value between {cat1} and {cat2}\n")
            print(f"Larger\t{str(pval1)}\n")
            print(f"Smaller\t{str(pval2)}\n")

        elif alt_hyp == "two-sided":
            stat, pval = proportion.proportions_ztest([s1, s2], [n1, n2],
                                                      alternative='two-sided')
            results.write(f"Two-sided test p-value between {cat1} and {cat2}\t"
                          f"{str(pval)}\n")
            print(f"Two-sided test p-value between {cat1} and {cat2}\t"
                  f"{str(pval)}\n")
        else:
            sys.exit("The alternative hypothesis \"alt_hyp\" has to be"
                     "\"one-sided\" or \"two-sided\"")
        pvals[(cat1, cat2)] = pval

    results.close()
    print(f"Results saved in {output_dir}{output_file}")
    return {"categories": cats, "proportions": prop,
            "confidence intervals": cim, "p-values": pvals}


def enrichment_test(dict_cats,
                    dict_features,
                    targ_features,
                    all_features="all",
                    targ_cats="all",
                    all_cats="all",
                    min_nb_elements=4,
                    output_dir=f"{resdir}enrichment_test/",
                    output_file=f"enrich_test{datetime.now()}"):
    '''
    Computes enrichment tests (hypergeometric test) of a target in a
    category.

    Args:
        dict_cats (dict.): classification of each element in a dictionary of
                shape {category:[elements]}
                N.B.: the same element can be associated to multiple features
        dict_features (dict.): feature corresponding to each element in a
                dictionary of shape {feature:[elements]}
        targ_features (list or string): targeted feature(s)
        all_features (Optional[list or str.]): features that will be used to 
                calculate the proportion, including targ_features
                (default: all keys of the dict_features dictionary)
        targ_cats (Optional[list]): list of categories. The enrichment test is 
                performed for each category. (default: all keys of dict_cats)
        all_cats (Optional[list]): list of categories used to compute the global
                proportion and the expected number in the selection,
                including targ_cats.
                (default: all keys of dict_cats)
        min_nb_elements (Optional[int.]): Number of elements used as thresholds 
                for the feature selection. If there is strictly less than 
                min_nb_elements corresponding to a feature, the result 
                corresponding to this feature is not relevant and is 
                therefore neither returned nor reported in the output file.
                By default, min_nb_elements is set to 4.
        output_dir (Optional[str.]): output directory
        output_file (Optional[str.]): output filename

    Returns: 
        DataFrame: DataFrame containing the following columns:
            * 'Category'(str.): category
            * 'Selected_gene_nb'(int.): Nb of elements corresponding to this
              feature in the selection
            * 'Expected_selected_nb' (int.): Expected nb of elements 
              corresponding to this feature in the selection
            * 'Total_gene_nb'(int.): Nb of elements corresponding to this
              feature in the dict_features
            * 'Proportion'(float): ratio between Selected_gene_nb and 
              Total_gene_nb
            * 'Prop_conf_int' (np.array): 95% confidence interval with equal
              areas around the proportion.
            * 'p-value' (float): p-value obtained with the enrichment test
            * 'Adj p-value (FDR)' (float): p-value corrected for false
              discovery rate
            * 'Global_proportion' (float): ratio between nb of elements in 
              the selection and nb of elements in dict_features

    Note: 
        This function performs a p-value correction for false discovery rate using
        statsmodels.stats.multitest.fdrcorrection 

    Note: 
        The created DataFrame, ordered according to the adjusted pvalues,
        is reported in the output_file.

    Example:
        >>> from GRATIOSA import stat_analysis
        >>> dict_features = {"act": ["B", "D", "E", "H", "I", "M", "P", "Q", "R", "S", "T", "W"], 
        ...                  "rep": ["C", "F", "G", "U", "X"], "None": ["A"], "NA": ["J"]}                
        >>> dict_cats = {"GOterm1": ["A", "B", "D", "E", "F", "P", "Q", "R", "S", "T", "U"], 
        ...              "GOterm2": ["C", "E"], 
        ...              "GOterm3": ["A", "B", "F", "G", "H", "I", "M", "U", "V", "W", "X"], 
        ...              "GOterm4": ["C", "F", "G", "J"]}               
        >>> stat_analysis.enrichment_test(dict_cats,
        ...                           dict_features,
        ...                           targ_features=["act", "None"],
        ...                           all_features=["act", "None", "rep", "NA"],
        ...                           targ_cats=["GOterm1", "GOterm2", "GOterm3"],
        ...                           min_nb_elements=3,
        ...                           output_file="test")
          Category  Selected_gene_nb  Total_gene_nb  Proportion  Prop_conf_int  p-value  
        0  GOterm1                9          11    0.818182  [0.4545, 1.0]  0.27206
        1  GOterm3                6          10    0.600000     [0.2, 1.0]  0.97059
          Adj p-value (FDR)  Global_proportion  Expected_selected_nb
        0           0.54412        0.722222        7.944444
        1           0.97059        0.722222         7.22222
          Category  Selected_gene_nb  Total_gene_nb  Proportion     Prop_conf_int 
        0  GOterm1                9          11    0.818182  [0.6818, 0.9545]
        1  GOterm3                6          10    0.600000        [0.4, 0.8]
          p-value  Adj p-value (FDR)  Global_proportion  Expected_selected_nb
        0 0.16563            0.33126        0.684211        7.526316
        1 0.90867            0.90867        0.684211        6.84210
        # GOterm2 was ignored because its nb of elements is less than 3.
        # GOterm4 was ignored because it was not selected in the "features"
    '''
    # Formats the objects given as arguments as lists containing no duplicates.
    if isinstance(targ_features, str):
        targ_features = [targ_features]
    else:
        targ_features = list(dict.fromkeys(targ_features))

    if isinstance(all_features, str):
        if all_features == "all":
            all_features = dict_features.keys()
    else:
        all_features = list(dict.fromkeys(all_features))

    if isinstance(targ_cats, str):
        if targ_cats == "all":
            targ_cats = list(dict_cats.keys())
        else:
            targ_cats = [targ_cats]
    else:
        targ_cats = list(dict.fromkeys(targ_cats))

    if isinstance(all_cats, str):
        if all_cats == "all":
            all_cats = list(dict_cats.keys())
    else:
        all_cats = list(dict.fromkeys(all_cats))

    if set(targ_cats) - set(all_cats):
        sys.exit("Every category listed in \"targ_cats\""
                 "has to be included in \"all_cats\".\n")

    # Keeps only valid elements ie elements that correspond:
    #   to a category listed in all_cats
    #   and to a feature listed in all_features
    cat_elements = set([el for c in all_cats for el in dict_cats[c]])
    all_targ_elements = set(
        [el for t in all_features for el in dict_features[t]])
    valid_elements = cat_elements & all_targ_elements
    M = len(valid_elements)    # M: total number of valid elements

    # Selects valid elements that correspond to a feature listed in
    # targ_features
    targ_elements = set(
        [el for tA in targ_features for el in dict_features[tA]])
    selection = valid_elements & targ_elements
    n = len(selection)    # n: Nb of valid elements in the selection
    global_average = n / M

    # Perfoms enrichment test for each category
    file = []
    for cat in targ_cats:
        # N: Nb of valid elements in the category
        N = len(set(dict_cats[cat]) & valid_elements)
        if len(targ_cats) < 20:
            print(f"{cat}: {N} valid elements")

        # x: Nb of valid elements associated to the category in the selection
        x = len(set(dict_cats[cat]) & selection)
        if N != 0:
            prop = x / N

            # exp_nb: expected number of elements associated to the category
            # in the selection
            exp_nb = global_average * N

            # Hypergeometric test (See scipy.stats.hypergeom documentation)
            pval = stats.hypergeom.sf(x - 1, M, n, N)

            # ci0 and ci1 are the 95% confidence interval with equal areas
            # around the proportion.
            ci = stats.hypergeom.interval(0.95, M, n, N, loc=0)
            diff_ci = (ci[1] - ci[0]) / float(N)
            ci0, ci1 = max(prop - diff_ci / 2, 0), min(prop + diff_ci / 2, 1)

            # Keeps relevant results i.e. categorys with more than
            # "min_nb_elements" elements
            if N >= min_nb_elements:
                file.append([cat, pval, x, N, prop,
                             np.round((ci0, ci1), 4), global_average, exp_nb])

    df = pd.DataFrame(data=file, columns=['Category', 'p-value', 'Selected_gene_nb',
                                          'Total_gene_nb', 'Proportion',
                                          'Prop_conf_int', 'Global_proportion',
                                          'Expected_selected_nb'])

    # pvalue correction for false discovery rate
    # (See statsmodels.stats.multitest.fdrcorrection documentation)
    df['Adj p-value (FDR)'] = multitest.fdrcorrection(df['p-value'])[1]

    # Returns and saves the results as a dataframe
    df = df[['Category', 'Selected_gene_nb', 'Total_gene_nb', 'Proportion', 'Prop_conf_int',
             'p-value', 'Adj p-value (FDR)', 'Global_proportion', 'Expected_selected_nb']]
    # Sorts the results according to the adjusted pvalues
    df = df.sort_values(by=['Adj p-value (FDR)'], ascending=True)

    Path(output_dir).mkdir(parents=True, exist_ok=True)
    df2csv = df.round({'Proportion': 3,
                       'Global_proportion': 3,
                       'p-value': 3,
                       'Adj p-value (FDR)': 3,
                       'Expected_selected_nb': 3})
    df2csv.to_csv(f"{output_dir}{output_file}.csv", sep='\t', index=False)
    print(f"Results saved in {output_dir}{output_file}.csv")

    return df


def quantitative_data_student_test(dict_data, cats="all",
                                   alt_hyp="one-sided",
                                   output_dir=f"{resdir}student_test/",
                                   output_file=f"student_test{datetime.now()}"):
    '''
    Computes the T-test for the means of independants categories using the
    scipy.stats.ttest_ind.

    Args:
        dict_data (dict.): datapoints corresponding to each category in a
                dictionary of shape {category:list of datapoints}
        cats (Optional [list]): list of categories to compare
                (default: all keys of dict_data)
        alt_hyp (Optional ["two-sided" or "one-sided"]): alternative hypothesis.
                If "one-sided" is chosen, both one-sided tests will be 
                performed with the scipy.stats.ttest_ind function and the 
                smaller p-value will be kept. See scipy documentation for 
                more details. (default: "one-sided")
        output_dir (Optional[str.]): output directory
        output_file (Optional[str.]): output filename

    Returns:
        Dictionary: dict. of shape 
        {"categories":cats, "means":means,
        "confidence intervals":(ci0,ci1), "p-values":{(cat1,cat2):pval}
    Note:
        Test results are also reported in the output_file.

    Example:
        >>> from GRATIOSA import stat_analysis
        >>> dict_data = {'a':[1,2,5,6,19], 'b':[10,24,4,15]}
        >>> stat_analysis.quantitative_data_student_test(dict_data)
        {'categories': ['a', 'b'],
         'means': [6.6, 13.25],
         'confidence intervals': [(0.2610995224360444, 12.938900477563955),
                                  (4.958672795504613, 21.54132720449539)],
         'p-values': {('a', 'b'): 0.12169488735182109}}
    '''
    if cats == "all":
        cats = list(dict_data.keys())
    elif isinstance(cats, str):
        cats = [cats]

    # Saves the results in a .txt file.
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    results = open(f'{output_dir}{output_file}.txt', 'w')
    results.write(f"Category\tMean\tConfidence interval\n")
    means, cim = [], []
    for c in cats:
        m = np.mean(dict_data[c])
        means.append(m)
        ci = stats.norm.interval(alpha=0.95, loc=m,
                                 scale=stats.sem(dict_data[c]))
        cim.append((ci[0], ci[1]))
        results.write(f"{c}\t{m}\t{(ci[0],ci[1])}\n")
    results.write("\n")

    # Tests pairwise if the means between categories are significantly
    # differents. Performs one-sided or two-sided proportions test depending 
    # on the "alt_hyp" argument.
    comp = []
    pvals = {}
    for i in np.arange(len(cats)):
        for j in np.arange(i + 1, len(cats)):
            comp.append([cats[i], cats[j]])
    for cat1, cat2 in comp:
        stats.ttest_ind(dict_data[cat1], dict_data[cat2])
        if alt_hyp == "one-sided":
            stat1, pval1 = stats.ttest_ind(dict_data[cat1], dict_data[cat2],
                                           alternative='greater')
            stat2, pval2 = stats.ttest_ind(dict_data[cat1], dict_data[cat2],
                                           alternative='less')
            pval = min(pval1, pval2)
            results.write(
                f"One-sided test\tp-value between {cat1} and {cat2}\n")
            results.write(f"Larger\t{str(pval1)}\n")
            results.write(f"Smaller\t{str(pval2)}\n")
        elif alt_hyp == "two-sided":
            stat, pval = stats.ttest_ind(dict_data[cat1], dict_data[cat2],
                                         alternative='two-sided')
            results.write(f"Two-sided test p-value between {cat1} and {cat2}\t"
                          f"{str(pval)}\n")
        else:
            sys.exit("The alternative hypothesis \"alt_hyp\" has to be"
                     "\"one-sided\" or \"two-sided\"")
        pvals[(cat1, cat2)] = pval
    results.close()
    print(f"Results saved in {output_dir}{output_file}")
    return {"categories": cats, "means": means,
            "confidence intervals": cim, "p-values": pvals}


#! /usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
import operator
import numpy as np
from datetime import datetime
from pathlib import Path
from GRATIOSA.useful_functions_genome import *
from GRATIOSA.globvar import *
from GRATIOSA.Gene import Gene
from GRATIOSA.TSS_TTS_TU import TSS, TTS, TU


class Genome:

    '''
    The Genome class is one of the main classes of this package. It gathers all
    the attributes of a genome such as the sequence, the set of genes (with their
    functional annotations and their orientation) but also all the annotations of
    TSS, TU and TTS.

    Each Genome instance has to be initialized with an organism name
        >>> from GRATIOSA import Genome
        >>> g = Genome.Genome("ecoli")
    '''

    def __init__(self, name):
        self.name = name

    def load_seq(self, filename="sequence.fasta"):
        """
        Load_seq loads DNA sequence from a .fasta file present in the
        main directory of the organism using useful_functions_genome.load_seq
        function. Adds this sequence, its complement, and its length to a
        Genome instance.
        
        Creates 3 new attributes to the Genome instance
            * seq (str.): genomic sequence compose of A,T,G and C
            * seqcompl (str.): complement sequence to seq
            * length (int.): length of the genomic sequence

        Args:
            filename (Optional [str.}): name of the file containing the DNA
                    sequence in FASTA format. Default: "sequence.fasta"

        Example:
            >>> from GRATIOSA import Genome
            >>> g = Genome.Genome("dickeya")
            >>> g.load_seq()
            >>> g.seq[100:151]
            'AATGTCGATCTTCAACATATCGCCGATCCGACGGGCACCCAGATCCTGCAG'
            >>> g.seqcompl[100:151]
            'TTACAGCTAGAAGTTGTATAGCGGCTAGGCTGCCCGTGGGTCTAGGACGTC'
        """
        seq_dir = f"{basedir}data/{self.name}/sequence.fasta"
        self.seq = read_seq(seq_dir).upper()
        self.seqcompl = ''
        self.length = len(self.seq)
        l = self.seq
        l = l.replace('A', 't').replace('T', 'a')
        l = l.replace('C', 'g').replace('G', 'c')
        l = l.replace('a', 'A').replace('t', 'T')
        l = l.replace('c', 'C').replace('g', 'G')
        self.seqcompl = l


    def load_annotation(self, annot_file="sequence.gff3"):
        """
        load_annotation loads a gene annotation (coordinates, length,
        name...) from a file present in the /annotation/ directory.

        Creates a new attribute of the Genome instance: 
            * self.genes (dict.) 
                self.genes is a dictionary of shape 
                {locus_tags: Gene object}. Each Gene object is 
                initialized with the following attributes:
                locus_tag, ID, name,strand, left, right, start, end, 
                middle, length and ASAP. 

        Args:
            filename (Optional [str.]): name of the file containing the 
                    genomic annotation. Default: "sequence.gff3"

        Note: 
            If the file is a .gff3 or .gff information will be loaded using
            useful_functions_genome.load_gff.
            Else, the information importation requires an annotation.info file,
            containing column indices of each information in the data file and 
            some additional information, in the following order:
            [0] Filename [1] Separator [2] Locus_tag column [3] Name column
            [4] ID column [5] Strand column [6] Left coordinate column
            [7] Right coordinate column [8] File start line

        Example:
            >>> from GRATIOSA import Genome
            >>> g = Genome.Genome("dickeya")
            >>> g.load_annotation()
            >>> g.genes["Dda3937_00005"].name
            'guaA'
        """
        path2dir = f"{basedir}data/{self.name}/annotation/"
        path2file = f"{path2dir}{annot_file}"
        file_type = Path(path2file).suffix

        print(f"Trying to load annotation from: {path2file}")
        if file_type in [".gff", ".gff3"]:
            self.genes = load_gff(path2file)
            load = True
        else:
            load = False
            with open(f"{path2dir}annotation.info", "r") as f:
                skiphead = next(f)
                for line in f:
                    line = line.strip().split('\t')
                    if line[0] == annot_file:
                        self.genes = load_annot_general(
                            f"{path2dir}{line[0]}",
                            line[1], int(line[2]),
                            int(line[3]), int(line[4]),
                            int(line[5]), int(line[6]),
                            int(line[7]), int(line[8]))
                        load = True
                f.close()
            if not load:
                print(f"Unable to load annotation from: {path2file}."
                      "Please add information in annotation.info or use gff3 or gff format.")
        if load == True:
            print("Annotation loaded")


    def load_genes_per_pos(self, window=0):
        """
        load_genes_per_pos associates each position with a list of genes
        overlapping this and its surrounding positions delimited
        by the window size given as an argument. 

        Creates a new attribute of the Genome instance: 
            * self.genes_per_pos (dict.)
                    self.genes is a dictionary of shape {position: list of 
                    genes}. It contains, for each position p, the list of
                    genes overlapping any position between p-window/2 and 
                    p+window/2 (inclusive)

        Args:
            window (Optional [int.]): window size in b. load_genes_per_pos 
                    finds genes between pos-window/2 and pos+window/2
                    (inclusive). Default: window = 0

        Warning:
            This method needs a genomic annotation. If no annotation is
            loaded, the load_annotation method with the default "sequence.gff3"
            file is computed. To use another annotation, please load an
            annotation before using the load_genes_per_pos method.

        Example:
            >>> from GRATIOSA import Genome
            >>> g = Genome.Genome("dickeya")
            >>> g.load_annotation(annot_file="sequence.gff3")
            >>> g.load_genes_per_pos(window=1000)
            >>> g.genes_per_pos[120]
            ['Dda3937_00155', 'Dda3937_00156', 'Dda3937_00154']
        """
        if not hasattr(self, "genes"):
            self.load_annotation()
        if not hasattr(self, "seq"):
            self.load_seq()

        self.genes_per_pos = {}

        for locus in self.genes.keys():
            gene = self.genes[locus]
            for pos in np.arange(gene.left - int(window / 2),
                                 gene.right + 1 + int(window / 2)):
                if pos < 0:
                    pos += self.length
                if pos > self.length:
                    pos -= self.length
                if pos in self.genes_per_pos.keys():
                    self.genes_per_pos[pos].append(locus)
                else:
                    self.genes_per_pos[pos] = [locus]

        for pos in np.arange(self.length):
            if pos not in self.genes_per_pos.keys():
                self.genes_per_pos[pos] = [None]


    def load_neighbor_all(self):
        """
        For each gene and positions, load_neighbor_all finds nearest
        neighbors (left and right) on genome, whatever their strand.
        
        Creates 2 new attributes of Gene instances:
            * self.genes[locus].left_neighbor 
                locus of the nearest left-side neighbor gene
            * self.genes[locus].right_neighbor 
                locus of the nearest right-side neighbor gene

        and 4 new attributes of Genome instance:
            * self.genomic_situation (dict.) 
                dict of shape {position: situation} with situation either 
                "intergenic" or "intragenic".
            * self.left_neighbor (dict.) 
                dict of shape {position: locus of the nearest left-side 
                neighbor gene}
            * self.right_neighbor (dict.) 
                dict of shape {position: locus of the nearest right-side 
                neighbor gene}
            * self.gene (dict.) 
                dict of shape {position: gene} with gene = "NA" if the 
                position is intergenic.
            
        Warning:
            This method needs a genomic annotation. If no annotation is 
            loaded, the load_annotation method with the default "sequence.gff3"
            file is computed. To use another annotation, please load an
            annotation before using the load_neighbor_all method.

        Example:
            >>> from GRATIOSA import Genome
            >>> g = Genome.Genome("dickeya")
            >>> g.load_annotation(annot_file="sequence.gff3")
            >>> g.load_neighbor_all()
            >>> g.genes["Dda3937_00005"].left_neighbor
            'Dda3937_00006'
            >>> g.genes["Dda3937_00005"].right_neighbor
            'Dda3937_00004'
            >>> g.g.genomic_situation[569]
            'intragenic'
            >>> g.gene[569]
            'Dda3937_00156'
            >>> g.left_neighbor[569]
            'Dda3937_00155'
            >>> g.right_neighbor[569]
            'Dda3937_00157'
        """
        if not hasattr(self, "genes"):
            self.load_annotation()
        if not hasattr(self, "length"):
            self.load_seq()

        # Sorts genes according to their position.
        list_start = []
        for i in self.genes:
            list_start.append(int(self.genes[i].start))
        genes_order = [x for _, x in sorted(
            zip(list_start, self.genes.keys()))]

        self.genomic_situation = ["intergenic"] * self.length
        self.left_neighbor = {}
        self.right_neighbor = {}
        self.gene = ["NA"] * self.length

        for i in np.arange(len(genes_order)):
            # Adds right and left neighbors to each Gene object
            if i != 0:
                self.genes[genes_order[i]].add_left_neighbor(
                    genes_order[i - 1])
            else:
                self.genes[genes_order[i]].add_left_neighbor(genes_order[-1])
            if i != len(genes_order) - 1:
                self.genes[genes_order[i]].add_right_neighbor(
                    genes_order[i + 1])
            else:
                self.genes[genes_order[i]].add_right_neighbor(genes_order[0])

            # Adds right and left neighbors to each genomic position
            for p in np.arange(self.genes[genes_order[i]].left, self.genes[genes_order[i]].right):
                self.genomic_situation[p] = "intragenic"
                self.gene[p] = genes_order[i]
                if i != 0:
                    self.left_neighbor[p] = genes_order[i - 1]
                else:
                    self.left_neighbor[p] = genes_order[-1]
                if i != len(genes_order) - 1:
                    self.right_neighbor[p] = genes_order[i + 1]
                else:
                    self.right_neighbor[p] = genes_order[0]
            if i != len(genes_order) - 1:
                for p in np.arange(self.genes[genes_order[i]].right, self.genes[genes_order[i + 1]].left):
                    self.left_neighbor[p] = genes_order[i]
                    self.right_neighbor[p] = genes_order[i + 1]
            else:
                for p in np.arange(self.genes[genes_order[i]].right, self.length):
                    self.left_neighbor[p] = genes_order[i]
                    self.right_neighbor[p] = genes_order[0]
                for p in np.arange(0, self.genes[genes_order[0]].left):
                    self.left_neighbor[p] = genes_order[i]
                    self.right_neighbor[p] = genes_order[0]


    def load_gene_orientation(self, couple=3, max_dist=5000):
        """
        Compute gene orientation with the following criteria: 

        * If couple = 3, gene is considered:

            * `divergent` if left neighbor on - strand and right neighbor on + strand,
            * `convergent` if left neighbor on + strand and right neighbor on - strand,
            * `tandem` if left and right neighbors on same strand (whatever the strand of the given gene is),
            * `isolated` if the distance between neighbors is higher than the maximal distance given as argument.
        * If couple = 2, gene is considered:

            * `tandem` if predecessor (left neighbor for gene on + strand, right neighbor for gene on - strand) is on same strand,
            * `divergent` if the predecessor is on opposite strand.

        Creates new attributes: 
            * self.orientation (dict.) 
                    new attribute of the Genome instance.
                    self.orientation is a dictionary of shape {orientation: 
                    list of genes}. It contains the list of genes for each
                    orientation (tandem, divergent, convergent, and isolated 
                    if couple=3, tandem and divergent if couple=2)
            * self.genes[locus].orientation (str.) 
                    new attribute of Gene instances related to the Genome instance 
                    given as argument

        Args:
            couple (int.): number of genes to consider in a "couple". 

                * If couple = 2: computes the orientation of a gene 
                  relative to its predecessor
                * If couple = 3: computes the orientation of a gene 
                  relative to its two neighbors

                Default: 3

            max_dist (Optional [int.]): maximal distance between 2 genes 
                    start positions for seeking neighbor (Default: 5kb)            

        Warning:
            This method needs a genomic annotation. If no annotation is 
            loaded, the load_annotation method with the default "sequence.gff3"
            file is computed. To use another annotation, please load an
            annotation before using the load_neighbor_all method.

        Example:
            >>> from GRATIOSA import Genome
            >>> g = Genome.Genome("dickeya")
            >>> g.load_annotation(annot_file="sequence.gff3")
            >>> g.load_gene_orientation()
            >>> g.genes["Dda3937_00005"].orientation
            'tandem'
            >>> g.orientation["isolated"]
            ['Dda3937_02360', 'Dda3937_01107', 'Dda3937_03898', 'Dda3937_04216',
            'Dda3937_00244', 'Dda3937_04441', 'Dda3937_01704', 'Dda3937_01705',
            ...
            'Dda3937_02126', 'Dda3937_01530', 'Dda3937_04419', 'Dda3937_02081']
        """
        self.load_neighbor_all()
        res = {"tandem": [], "divergent": [], "convergent": [], "isolated": []}
        for gene in self.genes:
            orient = ""
            try:
                g = self.genes[gene]
                lg = self.genes[g.left_neighbor]
                rg = self.genes[g.right_neighbor]

                # circular DNA
                if lg.start > g.start:
                    rstart = rg.start + self.length
                    gstart = g.start + self.length
                elif g.start > rg.start:
                    rstart = rg.start + self.length
                    gstart = g.start
                else:
                    rstart = rg.start
                    gstart = g.start

                if couple == 3:
                    if ((gstart - lg.start) < max_dist
                            and (rstart - gstart) < max_dist):
                        if not lg.strand and not rg.strand:
                            orient = "tandem"
                        elif lg.strand and rg.strand:
                            orient = "tandem"
                        elif lg.strand and not rg.strand:
                            orient = "convergent"
                        elif not lg.strand and rg.strand:
                            orient = "divergent"
                    else:
                        orient = "isolated"

                elif couple == 2:
                    if g.strand:
                        if (gstart - lg.start) < max_dist:
                            if lg.strand:
                                orient = "tandem"
                            elif not lg.strand:
                                orient = "divergent"
                    if not g.strand:
                        if (rstart - gstart) < max_dist:
                            if rg.strand:
                                orient = "divergent"
                            elif not rg.strand:
                                orient = "tandem"

                self.genes[gene].add_orientation(orient)
                res[orient].append(gene)
            except BaseException as e:
                print(f"Warning with locus {gene}: {e}")
        self.orientation = res


    def load_pos_orientation(self, max_dist=5000):
        """
        Computes gene orientation with the following criteria:

            * `divergent` if left neighbor on - strand and right neighbor on + strand,
            * `convergent` if left neighbor on + strand and right neighbor on - strand,
            * `tandem` if left and right neighbors on same strand
            * `isolated` if the distance between neighbors is higher than the maximal distance given as argument
        
        Creates 2 new attributes of the Genome instance:
            * self.pos_orientation (dict of dict)
                    dictionary containing 2 subdictionaries. 
                    One subdictionary for "intergenic" positions and one for 
                    "intragenic" positions.
                    Each subdictionary contains the list of position for 
                    each orientation. {"intergenic":{orientation: list of 
                    positions}}, "intragenic":{orientation: list of positions}}
                    with orientation in ["divergent","convergent","tandem",
                    "isolated"]
            * self.orientation_per_pos (dict.)
                    dictionary of shape {position: orientation}.

        Args:
            max_dist (Optional [int.]): maximal distance between 2 genes 
                    start positions for seeking neighbor (Default: 5kb)            

        Warning:
            This method needs a genomic annotation. If no annotation is
            loaded, the load_annotation method with the default "sequence.gff3"
            file is computed. To use another annotation, please load an
            annotation before using this method.

        Example:
            >>> from GRATIOSA import Genome
            >>> g = Genome.Genome("dickeya")
            >>> g.load_annotation(annot_file="sequence.gff3")
            >>> g.load_pos_orientation()
            >>> g.orientation_per_pos[30]
            'tandem'
            >>> g.pos_orientation["intragenic"]["isolated"]
            [11534,11535,11536,11537,11538,11539,11540,11541,11542,11543,...]
        """
        self.load_neighbor_all()
        res_inter = {
            "tandem": [],
            "divergent": [],
            "convergent": [],
            "isolated": []}
        res_intra = {
            "tandem": [],
            "divergent": [],
            "convergent": [],
            "isolated": []}
        self.orientation_per_pos = {}
        for pos in np.arange(self.length):
            orient = ""
            try:
                lg = self.genes[self.left_neighbor[pos]]
                rg = self.genes[self.right_neighbor[pos]]

                # circular DNA
                if lg.start > pos:
                    pstart = pos + self.length
                    rstart = rg.start + self.length
                elif pos > rg.start:
                    pstart = pos
                    rstart = rg.start + self.length
                else:
                    pstart = pos
                    rstart = rg.start

                if ((pstart - lg.start) < max_dist and (rstart - pstart) < max_dist):
                    if not lg.strand and not rg.strand:
                        orient = "tandem"
                    elif lg.strand and rg.strand:
                        orient = "tandem"
                    elif lg.strand and not rg.strand:
                        orient = "convergent"
                    elif not lg.strand and rg.strand:
                        orient = "divergent"
                else:
                    orient = "isolated"

                self.orientation_per_pos[pos] = orient
                if self.genomic_situation[pos] == "intragenic":
                    res_intra[orient].append(pos)
                else:
                    res_inter[orient].append(pos)
            except BaseException as e:
                print(f"Warning with position {pos}: {e}")
        self.pos_orientation = {
            "intragenic": res_intra,
            "intergenic": res_inter}


    def load_TSS(self):
        """
        load_TSS loads a TSS annotation from a file present in the /TSS/
        directory.

        Creates:
            * self.TSSs (dict. of dict.)
                    New attribute of the Genome instance
                    self.TSSs is a dictionary of shape 
                    {Condition: {TSS position: TSS object}}. 
                    One subdictionary is created for each condition listed in 
                    TSS.info file. Each TSS object is initialized with the 
                    following attributes: pos, genes, promoter, score, strand
                    The promoter attribute is a dictionary containing, for 
                    each sigma factor (keys) a subdictionary (value). The 
                    first created key of this subdictionary is "sites" and 
                    the associated value is a tuple containing the positions 
                    of promoter elements. See __init__ and add_promoter in 
                    the TSS class for more details about each attribute.

            * self.TSSs['all_TSSs'] (subdictionary of self.TSSs)
                    additional subdictionary of self.TSSs, of shape:
                    self.TSSs['all_TSSs']={TSSpos: [TSScond]} with [TSScond] 
                    the list of TSS conditions where this TSS was found.        

        Note:
            The information importation requires a TSS.info file,
            containing column indices of each information in the data file and 
            some additional information, in the following order:
            [0] Condition [1] Filename [2] Locus tag column [3] TSS position
            [4] File start line [5] Separator [6] Strand column
            [7] Sigma factor column [8] Sites column [9] Score column

        Note:
            If some of the data types are missing (locus tags, sigma factors,
            scores or sites), an empty space can be left in the .info file.

        Warning:
            This method needs a genomic annotation. If no annotation is
            loaded, the load_annotation method with the default "sequence.gff3"
            file is computed. To use another annotation, please load an
            annotation before using this method.

        Example:
            >>> from GRATIOSA import Genome
            >>> g = Genome.Genome("dickeya")
            >>> g.load_TSS()
            >>> g.TSSs["dickeya-btss"][4707030].promoter
            {'sigma70': {'sites': (4707039, 4707044, 4707060, 4707065)},
             'sigma32': {'sites': (4707037, 4707046, 4707062, 4707068)}}
            >>> g.TSSs["dickeya-btss"][4707030].strand
            False
        """
        self.TSSs = {}
        self.TSSs['all_TSS'] = {}
        if not hasattr(self, "genes"):
            self.load_annotation()

        path2dir = f"{basedir}data/{self.name}/TSS/"
        if Path(f"{path2dir}TSS.info").exists():
            with open(f"{path2dir}TSS.info", "r") as f:
                skiphead = next(f)
                for line in f:
                    line = line.strip('\n').split('\t')
                    filename = line[1]
                    startline = int(line[4])
                    sep = line[5]
                    strand = int(line[6])
                    TSScol = int(line[3])
                    genescol = int(line[2]) if line[2] != "" else None
                    sigcol = int(line[7]) if line[7] != "" else None
                    sitescol = int(line[8]) if line[8] != "" else None
                    scorecol = int(line[9]) if line[9] != "" else None
                    try:
                        self.TSSs[line[0]] = load_TSS_cond(self.genes,
                                                           path2dir + filename,
                                                           TSScol, startline,
                                                           sep, strand,
                                                           genescol, sigcol,
                                                           sitescol, scorecol)
                        # appends all entries to the "all_TSS" subdictionary
                        for TSSpos in self.TSSs[line[0]].keys():
                            if TSSpos in self.TSSs['all_TSS'].keys():
                                self.TSSs['all_TSS'][TSSpos].append(line[0])
                            else:
                                self.TSSs['all_TSS'][TSSpos] = [line[0]]
                    except Exception as e:
                        print("Error loading", line[0], e)
        else:
            print("No TSS.info, unable to load TSS")


    def load_prom_elements(self, shift=0, prom_region=[0, 0]):
        """
        load_prom_elements extracts sequences of the different promoter 
        elements (spacer, -10, -35, discriminator, region around TSS) based 
        on -35 and -10 coordinates loaded with load_TSS method, for all TSS 
        conditions. For each sigma factor associated to a TSS annotation, creates 
        a subdictionary in self.TSSs[condTSS][TSS].promoter with the shape
        promoter[sigma] = {element: sequence of the element} with element 
        in ["spacer", "minus10", "minus35", "discriminator", "region"].

        Args:
            shift (Optional [int.]): number of nucleotides to include beyond 
                    each region on either side (Default: 0nt)
            prom_region (Optional [int.,int.]): region upstream and 
                    downstream TSSs to extract. Argument with the shape:
                    [length before TSS, length after TSS]. 
                    Default: [0,0] ie no sequence will be extracted around 
                    TSS.
        Note:
            See load_TSS description to understand the structure of the
            subdictionary self.TSSs[condTSS][TSS].promoter

        Warning:
            This method needs a genomic annotation. If no annotation is
            loaded, the load_annotation method with the default "sequence.gff3"
            file is computed. To use another annotation, please load an
            annotation before using this method.

        Example:
            >>> from GRATIOSA import Genome
            >>> g = Genome.Genome("dickeya")
            >>> g.load_prom_elements()
            >>> g.TSSs["dickeya-btss"][4707030].promoter
            {'sigma70': {'sites': (4707039, 4707044, 4707060, 4707065),
                         'spacer': 'TCGCCCACCCTCAAT',
                         'minus10': 'CATCAT',
                         'minus35': 'TACCCC',
                         'discriminator': 'GAATAACC'},
             'sigma32': {'sites': (4707037, 4707046, 4707062, 4707068),
                         'spacer': 'CCTCGCCCACCCTCA',
                         'minus10': 'ATCATCATGA',
                         'minus35': 'CCGTACC',
                         'discriminator': 'ATAACC'}}
        """
        if not hasattr(self, 'TSSs'):
            self.load_TSS()
        if not hasattr(self, 'seq'):
            self.load_seq()
        for cond_TSS in self.TSSs.keys():
            try:
                if cond_TSS != 'all_TSS':
                    for TSS in self.TSSs[cond_TSS].keys():
                        self.TSSs[cond_TSS][TSS].add_prom_elements(
                            self.seq, self.seqcompl, shift=shift, prom=prom_region)

            except BaseException:
                print('Unable to load prom elements for:', cond_TSS)


    def load_TU(self):
        """
        load_TU loads a TU annotation from a file present in the /TU/
        directory.

        Creates a new attribute of the Genome instance:
            * self.TUs (dict. of dict.)
                    self.TUs is a dictionary of shape
                    {Condition: {TU start pos: TU object}}
                    One subdictionary is created for each condition listed 
                    in TU.info file. Each TU object is initialized with the 
                    following attributes:  start, stop, orientation, genes,
                    left,right, expression. 
                    See __init__ in the TU class for more details 
                    about each attribute.

        Note:
            The information importation requires a TU.info file,
            containing column indices of each information in the data file and 
            some additional information, in the following order:
            [0] Condition [1] Filename [2] TU ID [3] Start column [4] Stop column
            [5] Strand column [6] Gene column [7] File start line [8] Separator
            [9] Expression (optionnal)
            See the load_TU_cond function in useful_functions_genome for more
            details.

        Example:
            >>> from GRATIOSA import Genome
            >>> g = Genome.Genome("dickeya")
            >>> g.load_TU()
            >>> g.TUs["TU_Forquet"][2336702].right
            2339864
            >>> g.TUs["TU_Forquet"][2336702].strand
            True
        """
        self.TUs = {}
        path2dir = f"{basedir}data/{self.name}/TU/"
        if Path(f"{path2dir}TU.info").exists():
            with open(f"{path2dir}TU.info", "r") as f:
                skiphead = next(f)
                for line in f:
                    line = line.strip().split('\t')
                    try:
                        exprcol = None
                        if len(line) > 9 :
                            try:
                                exprcol = int(line[9])
                            except:
                                pass 
                        self.TUs[line[0]] = load_TU_cond(path2dir + line[1],
                                                         int(line[2]), 
                                                         int(line[3]),
                                                         int(line[4]), 
                                                         int(line[5]),
                                                         int(line[6]), 
                                                         int(line[7]),
                                                         line[8],
                                                         exprcol)
                    except BaseException:
                        print("Error loading cond", line[0])
            f.close()
        else:
            print("No TU.info file, please create one")


    def load_TTS(self):
        """
        load_TTS loads a TTS annotation from a file present in the /TTS/
        directory.

        Creates:
            * self.TTSs (dict. of dict.)
                New attribute of the Genome instance
                self.TTSs is a dictionary of shape
                {Condition: {TTS position: TTS object}}.
                One subdictionary is created for each condition listed 
                in TTS.info file. Each TTS object is initialized with the 
                following attributes: left, right,  start, end, strand,                    
                rho_dpdt, genes, seq, score.
                If the data do not contain information about associated 
                genes, sequence, or rho dependency, the corresponding 
                attributes will be initialized as "None".
                See __init__ in the TTS class for more details about 
                each attribute.
            * self.TTSs['all_TTSs'] (subdictionary of self.TTSs)
                additional subdictionary of self.TTSs, of shape:
                self.TTSs['all_TTSs']={TTSpos: [TTScond]}
                with [TTScond] the list of TTS conditions where this TTS 
                was found.

        Note: 
            The information importation requires a TTS.info file,
            containing column indices of each information in the data file and
            some additional information, in the following order:
            [0] Condition [1] Filename [2] Left coordinate column
            [3] Right coordinate column [4] Strand column [5] Startline
            [6] Separator [7] Sequence column
            and optionally:
            [8] Score column [9] Genes column [10] Rho dependency column

        Note: 
            See the load_TTS_cond function in useful_functions_genome for more
            details.

        Example:
            >>> from GRATIOSA import Genome
            >>> g = Genome.Genome("dickeya")
            >>> g.load_TTS()
            >>> g.TTSs["RhoTerm"][2791688].left
            2791688
        """
        self.TTSs = {}

        path2dir = f"{basedir}data/{self.name}/TTS/"
        if Path(f"{path2dir}TTS.info").exists():
            with open(f"{path2dir}TTS.info", "r") as f:
                skiphead = next(f)
                for line in f:
                    line = line.strip().split('\t')
                    seqcol = int(line[7]) if line[7] != "" else None
                    scorecol = int(line[8]) if line[8] != "" else None
                    genescol = int(line[9]) if line[9] != "" else None
                    try:
                        self.TTSs[line[0]] = load_TTS_cond(
                            path2dir + line[1],
                            line[6], int(line[5]),
                            int(line[2]), int(line[3]),
                            int(line[4]), int(line[10]),
                            seqcol, scorecol, genescol)
                    except Exception as e:
                        print("Error loading cond", line[0])
            f.close()
            self.TTSs["all"] = {}
            for x in self.TTSs.keys():
                self.TTSs["all"].update(self.TTSs[x])
        else:
            print("No TTS.info file, please create one")


    def load_GO(self):
        """
        load_GO loads file specified in GO.info to assign GO terms to genes.

        Creates:      
            * self.GO (dict. of dict)
                new attribute of the Genome instance. self.GO is a dictionary of 
                shape {annot_syst: {GOterm: list of genes}} i.e. one subdictionary 
                is created for each annotation system (such as GOc, COG or domain) 
                listed in GO.info file.
            * self.genes[locus].GO (list)
                new attribute of Gene instances related to the Genome instance given
                as argument. List of terms (such as GO terms) associated to the gene.            
        
        Note:           
            The information importation requires a GO.info file,
            containing column indices of each information in the data file and
            some additional information, in the following order:
            [0] Annotation system [1] Filename [2] Locus tag column
            [3] GOterm column [4] Separator
            GO.info and the data files have to be in the /GO/ directory

        Note:
            Other annotation systems such as COG, or domain assignment can also
            be used.

        Warning:
            This method needs a genomic annotation. If no annotation is
            loaded, the load_annotation method with the default "sequence.gff3"
            file is computed. To use another annotation, please load an
            annotation before using this method.

        Example:
            >>> from GRATIOSA import Genome
            >>> g = Genome.Genome("dickeya")
            >>> g.load_GO()
            >>> g.GO['GO']['GO:0000100']
            ['Dda3937_01944', 'Dda3937_03618']
            >>> g.genes["Dda3937_00004"].GO
            {'GO': ['GO:0005829','GO:0055114','GO:0046872','GO:0042802',
                    'GO:0000166','GO:0006177','GO:0003938','GO:0005829',
                    'GO:0055114','GO:0046872','GO:0042802','GO:0000166',
                    'GO:0006177','GO:0003938']}
        """
        if not hasattr(self, "genes"):
            self.load_annotation()

        warn_locus = []
        self.GO = {}
        path2dir = f"{basedir}data/{self.name}/GO/"

        # Opens .info file to obtain the information required to correctly
        # load the data for each annotation system.
        if Path(f"{path2dir}GO.info").exists():
            with open(f"{path2dir}GO.info", "r") as info_file:
                skiphead = next(info_file)
                for info_line in info_file:
                    info_line = info_line.strip().split('\t')
                    system = info_line[0]
                    filename = info_line[1]
                    tagtype = info_line[2]
                    tagcol = int(info_line[3])
                    GOcol = int(info_line[4])
                    separator = info_line[5]
                    startline = int(info_line[6])

                    # Loads the data: associates GO terms to each gene
                    print(f"Loading {system}...")
                    GOterms = []
                    with open(path2dir + filename, 'r') as GO_file:
                        dict_GO = {}
                        i = 0
                        while i < startline:
                            header = next(GO_file)
                            i += 1
                        for line in GO_file:
                            line = line.strip('\n')
                            if separator == '\\t':
                                line = line.split('\t')
                            else:
                                line = line.split(separator)
                            GOterms.append(line[GOcol])
                            if line[tagcol] in dict_GO.keys():
                                dict_GO[line[tagcol]].append(line[GOcol])
                            else:
                                dict_GO[line[tagcol]] = [line[GOcol]]

                    self.GO[system] = {}
                    GOterms = list(set(GOterms))
                    for t in GOterms:
                        self.GO[system][t] = []

                    for locus in self.genes.keys():
                        g = self.genes[locus]
                        try:
                            if tagtype == "ASAP":
                                name = g.ASAP
                            elif tagtype == "locus_tag":
                                name = locus
                            else:
                                sys.exit("Unknown \"tagtype\" in GO.info file:"
                                         "accepted tagtypes are ASAP and locus_tag")
                            self.genes[locus].add_GO(system, dict_GO[name])
                            for t in dict_GO[name]:
                                if locus not in self.GO[system][t]:
                                    self.GO[system][t].append(locus)
                        except Exception:
                            warn_locus.append(locus)

                    GO_file.close()
                    # Locuses associated with no GO term are listed in warn_locus and are
                    # printed as a warning for each annotation system
                    success = len(self.genes.keys()) - len(warn_locus)
                    print(f"\t{success} genes were successfully associated with some GO terms")
                    if len(warn_locus) > 20:
                        print(f"\t{len(warn_locus)} genes were not associated with any GO term")
                    else:
                        print(f"\t{warn_locus} are associated with no GO term")
            info_file.close()
        else:
            print("No GO.info file, please create one")



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from GRATIOSA.globvar import *
from datetime import datetime
from GRATIOSA import stat_analysis
import os
import re
import pandas as pd

class GO:
    '''
    The GO (Gene Ontology) class contains the information (names, definitions) 
    of the GO terms of the different categories (biological process, molecular 
    function and cellular component). The associated function 
    (GO_enrichment_test) allows the study of functional enrichment.
    '''

    def __init__(self,
                 filename="go-basic.obo",
                 path2dir=f"{basedir}data/",
                 obo_reload=False):
        """
        Args:
            filename (str.): name of the Gene Ontology .obo file
            path2dir (str.): path to the directory containing the .obo file
            obo_reload (Bool.) : if true, the Gene Ontology basic.obo file is
                    reloaded. This file allows each GO term to be associated 
                    with a category (GOp - biological_process, GOf - 
                    molecular_function or GOc - cellular_component), a name 
                    and a description

        Initializes GO object with the following attributes:
            dict_GO (dict. of dict.) 
                    dictionary containing one subdictionary
                    per GO term. Each subdictionary contains the "Name", the 
                    "Namespace" (biological_process, molecular_function or 
                    cellular_component) and the definition corresponding to 
                    the GO term.
            dict_GOc (dict. of dict.) 
                    dictionary containing only the GOterms
                    from "cellular_component" namespace. This dictionary 
                    contains one subdictionary per GO term, with the 
                    following shape self.dict_GOc["GOterm1"] = {"Name": GO 
                    name, "Definition": associated definition}
            dict_GOf (dict. of dict.) 
                    idem for the GOterms from "molecular_function" namespace
            dict_GOp (dict. of dict.) 
                    idem for the GOterms from "biological_process" namespace

        Example:
            >>> from GRATIOSA import GO
            >>> go = GO.GO()
            >>> go.dict_GOc['GO:0000015']
            {'Name': 'phosphopyruvate hydratase complex',
             'Definition': 'A multimeric enzyme complex that catalyzes the 
             conversion of 2-phospho-D-glycerate to phosphoenolpyruvate and 
             water. [GOC:jl, ISBN:0198506732]'}
        """
        if obo_reload:
            os.system(f"wget http://purl.obolibrary.org/obo/go/go-basic.obo \
                -P {basedir}data/")

        self.dict_GO = {}
        with open(path2dir + filename, "r") as f:
            for line in f:
                line = line.strip("\n").replace("\"", "").split(": ")
                if "id" in line:
                    if "GO" in line[1]:
                        ID = line[1]
                    else:
                        ID = "other"
                    self.dict_GO[ID] = {}
                elif "name" in line:
                    self.dict_GO[ID]["Name"] = line[1]
                    # print(self.dict_GO)
                elif "namespace" in line:
                    self.dict_GO[ID]["Namespace"] = line[1]
                elif "def" in line:
                    self.dict_GO[ID]["Definition"] = line[1]

        if 'other' in self.dict_GO.keys():
            self.dict_GO.pop('other')

        self.dict_GOc, self.dict_GOp, self.dict_GOf = {}, {}, {}
        for GO in self.dict_GO.keys():
            if self.dict_GO[GO]["Namespace"] == "biological_process":
                self.dict_GOp[GO] = {"Name": self.dict_GO[GO]["Name"],
                                     "Definition": self.dict_GO[GO]["Definition"]}
            elif self.dict_GO[GO]["Namespace"] == "molecular_function":
                self.dict_GOf[GO] = {"Name": self.dict_GO[GO]["Name"],
                                     "Definition": self.dict_GO[GO]["Definition"]}
            elif self.dict_GO[GO]["Namespace"] == "cellular_component":
                self.dict_GOc[GO] = {"Name": self.dict_GO[GO]["Name"],
                                     "Definition": self.dict_GO[GO]["Definition"]}


def GO_enrichment_test(dict_GOterms,
                       dict_features,
                       targ_features,
                       all_features="all",
                       targ_cats="all",
                       all_cats="all",
                       min_nb_elements=4,
                       thresh_padj=0.1,
                       output_dir=f"{resdir}enrichment_test/",
                       output_file=f"enrich_test{datetime.now()}",
                       obo_reload=False,
                       GO_filename="go-basic.obo",
                       GO_path2dir=f"{basedir}data/"):
    '''
    Computes enrichment tests (hypergeometric test) of a target in a
    category.
    N.B.: performs a p-value correction for false discovery rate (See
          statsmodels.stats.multitest.fdrcorrection for more details)

    Required args:
        dict_GOterms (dict.): classification of each element in a dictionary 
                of shape {category:[elements]}. N.B.: the same element can be 
                associated to multiple features
                Example: {"GOterm1":["GeneA","GeneB"],
                          "GOterm2:["GeneA","GeneC"]}
        dict_features (dict.): feature corresponding to each element in a
                dictionary of shape {feature:[elements]}
                Example: {"act": ["GeneA","GeneC","GeneD"],"rep":["GeneB"]}
        targ_features (list or string): targeted feature(s)

    Optional args:
        all_features (list or string): features that will be used to calculate
                the proportion, including targ_features (default: all keys of 
                the dict_features dictionary)
        targ_cats (list): list of categories. The enrichment test is 
                performed for each category. (default: all keys of dict_cats)
        all_cats (list): list of categories used to compute the global
                proportion and the expected number in the selection, 
                including targ_cats.(default: all keys of dict_cats)
        min_nb_elements: Number of elements used as thresholds for the feature
                selection. If there is strictly less than min_nb_elements
                corresponding to a feature, the result corresponding to this 
                feature is not relevant and is therefore neither returned 
                nor reported in the output file.
                By default, min_nb_elements is set to 4.
        thresh_padj (float.) : threshold on the adjusted p-value, used to 
                create the file containing only the significantly enriched
                GO terms
        output_dir (str.): output directory
        output_file (str.): output filename
        GO_filename (str.): name of the Gene Ontology .obo file
        GO_path2dir (str.): path to the directory containing the .obo file
        obo_reload (Bool.) : if true, the Gene Ontology basic.obo file is 
                reloaded. This file allows each GO term to be associated with 
                a category (GOf, GOp or GOc), a name and a description
    Returns:
        3 files (one for GOf, one for GOp and one for GOc) containing the
        following columns:
            0. 'Term'(str.): GO term
            1. 'Selected_gene_nb'(int.): Nb of elements corresponding to this
                    feature in the selection
            2. 'Expected_selected_nb' (int.): Expected nb of elements 
                    corresponding to this feature in the selection
            3. 'Total_gene_nb'(int.): Nb of elements corresponding to this
                    feature in the dict_features
            4. 'Proportion'(float): ratio between Selected_gene_nb and 
                    Total_gene_nb
            5. 'Prop_conf_int' (np.array): 95% confidence interval with equal
                    areas around the proportion.
            6. 'p-value' (float): p-value obtained with the enrichment test
            7. 'Adj p-value (FDR)' (float): p-value corrected for false
                    discovery rate
            8. 'Global_proportion' (float): ratio between nb of elements in 
                    the selection and nb of elements in dict_features
            The data are ordered according to the adjusted pvalues.
        1 common file containing the significantly enriched GO terms

    Example:
        >>> from GRATIOSA import Genome, GO
        >>> g = Genome.Genome("ecoli")
        >>> g.load_GO()
        >>> g.load_state_from_FC()
        >>> GO.GO_enrichment_test(g.GO['GO'],g.statesFC['Blot'],'act')
    '''
    # Formats the objects given as arguments as lists containing no duplicates.
    dGO = GO(filename=GO_filename, path2dir=GO_path2dir, obo_reload=obo_reload)

    list_GO = ["GOc", "GOp", "GOf"]

    dGOp, dGOc, dGOf = {}, {}, {}

    unknown_terms = []
    for t in dict_GOterms.keys():
        try:
            sp = dGO.dict_GO[t]['Namespace']
            if sp == "biological_process":
                dGOp[t] = dict_GOterms[t]
            elif sp == "molecular_function":
                dGOf[t] = dict_GOterms[t]
            elif sp == "cellular_component":
                dGOc[t] = dict_GOterms[t]

        except BaseException:
            unknown_terms.append(t)
    if len(unknown_terms) != 0:
        print(f"{len(unknown_terms)} GO terms are not in the GO annotation")

    for n, dict_cats in enumerate([dGOc, dGOp, dGOf]):
        o_file = f"{output_file}_{list_GO[n]}"
        df_res = pd.DataFrame(stat_analysis.enrichment_test(dict_cats=dict_cats,
                                                            dict_features=dict_features,
                                                            targ_features=targ_features,
                                                            all_features=all_features,
                                                            targ_cats=targ_cats,
                                                            all_cats=all_cats,
                                                            min_nb_elements=min_nb_elements,
                                                            output_dir=output_dir,
                                                            output_file=o_file))
        terms = df_res["Category"]
        names = []
        defs = []
        for t in terms:
            names.append(dGO.dict_GO[t]['Name'])
            defs.append(dGO.dict_GO[t]['Definition'])
        df2 = df_res.assign(name=names, definition=defs)

        df2[['p-value',
             'Adj p-value (FDR)']] = df2[['p-value',
                                          'Adj p-value (FDR)']].applymap(lambda x: '{:.2E}'.format(x))
        df2[['Proportion',
             'Global_proportion',
             'Expected_selected_nb']] = df2[['Proportion',
                                             'Global_proportion',
                                             'Expected_selected_nb']].applymap(lambda x: '{:.3f}'.format(x))
        df2csv = df2
        df2csv.to_csv(f"{output_dir}{o_file}.csv", sep='\t', index=False)

    f_out = open(f"{output_dir}{output_file}_padj{thresh_padj}.csv", 'w')
    f_out.write(f"GOtype\tCategory\tSelected_gene_nb\tTotal_gene_nb\tExpected_selected_nb\tp-value\tAdj p-value (FDR)\tName\tDefinition\n")

    for GOtype in ["GOc", "GOf", "GOp"]:
        empty_test = True
        with open(f"{output_dir}{output_file}_{GOtype}.csv", "r") as f:
            header = next(f)
            for line in f:
                line = line.strip('\n').split('\t')
                if float(line[6]) < thresh_padj:
                    f_out.write(f"{GOtype}\t{line[0]}\t{line[1]}\t{line[2]}\t{line[8]}\t{line[5]}\t{line[6]}\t{line[9]}\t{line[10]}\n")
                    empty_test = False
        f.close()
        if empty_test:
            f_out.write(f"{GOtype}\tNA\n")
    f_out.close()
    print(f"Results saved in {output_dir}{output_file}_significant.csv")


def gaf2annot(path2file, startline, regexp_locus):
    '''
    converts a .gaf file into an .annot file which can then be used to load
    GO information with the load_go method of the genome class.

    Args:
        path2file (str.): path to the file containing the data
        startline (int.): file start line
        regexp_locus (str.): Regular expression that matches only the
            locus in the gaf file
    Output:
        .annot file saved in the folder of the initial .gaf file.
    Example:
    >>> from GRATIOSA import GO
    >>> GO.gaf2annot("path2file/ecocyc.gaf",35,"b\\d\\d\\d\\d")
    '''
    if ".gaf" in path2file:
        path2file = path2file[:-4]

    f_out = open(path2file + ".annot", "w")
    with open(path2file + ".gaf", "r") as f_in:
        i = 0
        while i < startline:
            header = next(f_in)
            i += 1
        for line in f_in:
            line = line.strip('\n').split('\t')
            try:
                GOterm = line[4]
                locus_part = line[10].split('|')
                for l in locus_part:
                    if re.compile(regexp_locus).search(l):
                        f_out.write(f"{l}\t{GOterm}\n")
            except Exception as e:
                print(e)
    f_in.close()
    f_out.close()
